@@ -93,6 +93,9 @@
  * \333          - REP prefix (0xF3 byte) used as opcode extension.
  * \334          - LOCK prefix used instead of REX.R
  * \335          - disassemble a rep (0xF3 byte) prefix as repe not rep.
+ * \336          - force a REP(E) prefix (0xF2) even if not specified.
+ * \337		 - force a REPNE prefix (0xF3) even if not specified.
+ *                 \336-\337 are still listed as prefixes in the disassembler.
  * \340          - reserve <operand 0> bytes of uninitialized storage.
  *                 Operand 0 had better be a segmentless constant.
  * \360		 - no SSE prefix (== \364\331)
@@ -143,7 +146,7 @@ static int32_t regflag(const operand *);
 static int32_t regval(const operand *);
 static int rexflags(int, int32_t, int);
 static int op_rexflags(const operand *, int);
-static ea *process_ea(operand *, ea *, int, int, int, int32_t, int);
+static ea *process_ea(operand *, ea *, int, int, int, int32_t);
 static void add_asp(insn *, int);
 
 static int has_prefix(insn * ins, enum prefix_pos pos, enum prefixes prefix)
@@ -246,13 +249,10 @@ static int jmp_match(int32_t segment, int64_t offset, int bits,
 
     if (c != 0370 && c != 0371)
         return 0;
-    if (ins->oprs[0].opflags & OPFLAG_FORWARD) {
-        if ((optimizing < 0 || (ins->oprs[0].type & STRICT))
-            && c == 0370)
-            return 1;
-        else
-            return (pass0 == 0);        /* match a forward reference */
-    }
+    if ((optimizing <= 0 || (ins->oprs[0].type & STRICT)))
+         {
+            return 0;
+         }
     isize = calcsize(segment, offset, bits, ins, code);
     if (ins->oprs[0].segment != segment)
         return 0;
@@ -335,7 +335,8 @@ int64_t assemble(int32_t segment, int64_t offset, int bits, uint32_t cp,
                         out(offset, segment, &e->offset,
                             OUT_ADDRESS, wsize, e->segment, e->wrt);
                     offset += wsize;
-                } else if (e->type == EOT_DB_STRING) {
+                } else if (e->type == EOT_DB_STRING ||
+			   e->type == EOT_DB_STRING_FREE) {
                     int align;
 
                     out(offset, segment, e->stringval,
@@ -365,15 +366,8 @@ int64_t assemble(int32_t segment, int64_t offset, int bits, uint32_t cp,
     }
 
     if (instruction->opcode == I_INCBIN) {
-        static char fname[FILENAME_MAX];
+        const char *fname = instruction->eops->stringval;
         FILE *fp;
-        int32_t len;
-
-        len = FILENAME_MAX - 1;
-        if (len > instruction->eops->stringlen)
-            len = instruction->eops->stringlen;
-        strncpy(fname, instruction->eops->stringval, len);
-        fname[len] = '\0';
 
 	fp = fopen(fname, "rb");
 	if (!fp) {
@@ -383,17 +377,18 @@ int64_t assemble(int32_t segment, int64_t offset, int bits, uint32_t cp,
             error(ERR_NONFATAL, "`incbin': unable to seek on file `%s'",
                   fname);
 	} else {
-            static char buf[2048];
-            int32_t t = instruction->times;
-            int32_t base = 0;
+            static char buf[4096];
+            size_t t = instruction->times;
+            size_t base = 0;
+	    size_t len;
 
             len = ftell(fp);
             if (instruction->eops->next) {
                 base = instruction->eops->next->offset;
                 len -= base;
                 if (instruction->eops->next->next &&
-                    len > instruction->eops->next->next->offset)
-                    len = instruction->eops->next->next->offset;
+                    len > (size_t)instruction->eops->next->next->offset)
+                    len = (size_t)instruction->eops->next->next->offset;
             }
             /*
              * Dummy call to list->output to give the offset to the
@@ -402,7 +397,7 @@ int64_t assemble(int32_t segment, int64_t offset, int bits, uint32_t cp,
             list->output(offset, NULL, OUT_RAWDATA, 0);
             list->uplevel(LIST_INCBIN);
             while (t--) {
-                int32_t l;
+                size_t l;
 
                 fseek(fp, base, SEEK_SET);
                 l = len;
@@ -660,7 +655,8 @@ int64_t insn_size(int32_t segment, int64_t offset, int bits, uint32_t cp,
             osize = 0;
             if (e->type == EOT_DB_NUMBER)
                 osize = 1;
-            else if (e->type == EOT_DB_STRING)
+            else if (e->type == EOT_DB_STRING ||
+		     e->type == EOT_DB_STRING_FREE)
                 osize = e->stringlen;
 
             align = (-osize) % wsize;
@@ -672,16 +668,10 @@ int64_t insn_size(int32_t segment, int64_t offset, int bits, uint32_t cp,
     }
 
     if (instruction->opcode == I_INCBIN) {
-        char fname[FILENAME_MAX];
+	const char *fname = instruction->eops->stringval;
         FILE *fp;
-        int32_t len;
+        size_t len;
 
-        len = FILENAME_MAX - 1;
-        if (len > instruction->eops->stringlen)
-            len = instruction->eops->stringlen;
-        strncpy(fname, instruction->eops->stringval, len);
-        fname[len] = '\0';
-	
 	fp = fopen(fname, "rb");
 	if (!fp)
             error(ERR_NONFATAL, "`incbin': unable to open file `%s'",
@@ -695,8 +685,8 @@ int64_t insn_size(int32_t segment, int64_t offset, int bits, uint32_t cp,
             if (instruction->eops->next) {
                 len -= instruction->eops->next->offset;
                 if (instruction->eops->next->next &&
-                    len > instruction->eops->next->next->offset) {
-                    len = instruction->eops->next->next->offset;
+                    len > (size_t)instruction->eops->next->next->offset) {
+                    len = (size_t)instruction->eops->next->next->offset;
                 }
             }
             return instruction->times * len;
@@ -754,49 +744,48 @@ int64_t insn_size(int32_t segment, int64_t offset, int bits, uint32_t cp,
     return -1;                  /* didn't match any instruction */
 }
 
-static bool possible_sbyte(insn * ins, int op)
+static bool possible_sbyte(operand *o)
 {
-    return !(ins->forw_ref && ins->oprs[op].opflags) &&
-        optimizing >= 0 &&
-        !(ins->oprs[op].type & STRICT) &&
-        ins->oprs[op].wrt == NO_SEG && ins->oprs[op].segment == NO_SEG;
+    return !(o->opflags & OPFLAG_FORWARD) &&
+	optimizing >= 0 && !(o->type & STRICT) &&
+	o->wrt == NO_SEG && o->segment == NO_SEG;
 }
 
 /* check that opn[op]  is a signed byte of size 16 or 32 */
-static bool is_sbyte16(insn * ins, int op)
+static bool is_sbyte16(operand *o)
 {
     int16_t v;
 
-    if (!possible_sbyte(ins, op))
+    if (!possible_sbyte(o))
 	return false;
 
-    v = ins->oprs[op].offset;
+    v = o->offset;
     return v >= -128 && v <= 127;
 }
 
-static bool is_sbyte32(insn * ins, int op)
+static bool is_sbyte32(operand *o)
 {
     int32_t v;
 
-    if (!possible_sbyte(ins, op))
+    if (!possible_sbyte(o))
 	return false;
 
-    v = ins->oprs[op].offset;
+    v = o->offset;
     return v >= -128 && v <= 127;
 }
 
 /* check that  opn[op]  is a signed byte of size 32; warn if this is not
    the original value when extended to 64 bits */
-static bool is_sbyte64(insn * ins, int op)
+static bool is_sbyte64(operand *o)
 {
     int64_t v64;
     int32_t v32;
 
     /* dead in the water on forward reference or External */
-    if (!possible_sbyte(ins, op))
+    if (!possible_sbyte(o))
 	return false;
 
-    v64 = ins->oprs[op].offset;
+    v64 = o->offset;
     v32 = (int32_t)v64;
 
     warn_overflow(32, v64);
@@ -930,7 +919,7 @@ static int64_t calcsize(int32_t segment, int64_t offset, int bits,
         case 0141:
         case 0142:
 	case 0143:
-            length += is_sbyte16(ins, c & 3) ? 1 : 2;
+            length += is_sbyte16(opx) ? 1 : 2;
             break;
         case 0144:
         case 0145:
@@ -943,7 +932,7 @@ static int64_t calcsize(int32_t segment, int64_t offset, int bits,
         case 0151:
         case 0152:
         case 0153:
-            length += is_sbyte32(ins, c & 3) ? 1 : 4;
+            length += is_sbyte32(opx) ? 1 : 4;
             break;
         case 0154:
         case 0155:
@@ -980,20 +969,18 @@ static int64_t calcsize(int32_t segment, int64_t offset, int bits,
         case 0251:
         case 0252:
         case 0253:
-            length += is_sbyte64(ins, c & 3) ? 1 : 4;
+            length += is_sbyte64(opx) ? 1 : 4;
             break;
 	case 0260:
 	case 0261:
 	case 0262:
 	case 0263:
-	    length += 2;
 	    ins->rex |= REX_V;
 	    ins->drexdst = regval(opx);
 	    ins->vex_m = *codes++;
 	    ins->vex_wlp = *codes++;
 	    break;
 	case 0270:
-	    length += 2;
 	    ins->rex |= REX_V;
 	    ins->drexdst = 0;
 	    ins->vex_m = *codes++;
@@ -1052,6 +1039,14 @@ static int64_t calcsize(int32_t segment, int64_t offset, int bits,
 	    break;
         case 0335:
 	    break;
+	case 0336:
+	    if (!ins->prefixes[PPS_LREP])
+		ins->prefixes[PPS_LREP] = P_REP;
+	    break;
+	case 0337:
+	    if (!ins->prefixes[PPS_LREP])
+		ins->prefixes[PPS_LREP] = P_REPNE;
+	    break;
         case 0340:
             if (ins->oprs[0].segment != NO_SEG)
                 errfunc(ERR_NONFATAL, "attempt to reserve non-constant"
@@ -1098,7 +1093,7 @@ static int64_t calcsize(int32_t segment, int64_t offset, int bits,
 
                 if (!process_ea
                     (&ins->oprs[(c >> 3) & 7], &ea_data, bits,
-		     ins->addr_size, rfield, rflags, ins->forw_ref)) {
+		     ins->addr_size, rfield, rflags)) {
                     errfunc(ERR_NONFATAL, "invalid effective address");
                     return -1;
                 } else {
@@ -1474,7 +1469,7 @@ static void gencode(int32_t segment, int64_t offset, int bits,
         case 0142:
 	case 0143:
             data = opx->offset;
-            if (is_sbyte16(ins, c & 3)) {
+            if (is_sbyte16(opx)) {
                 bytes[0] = data;
                 out(offset, segment, bytes, OUT_RAWDATA, 1, NO_SEG,
                     NO_SEG);
@@ -1495,7 +1490,7 @@ static void gencode(int32_t segment, int64_t offset, int bits,
 	case 0147:
 	    EMIT_REX();
             bytes[0] = *codes++;
-            if (is_sbyte16(ins, c & 3))
+            if (is_sbyte16(opx))
                 bytes[0] |= 2;  /* s-bit */
             out(offset, segment, bytes, OUT_RAWDATA, 1, NO_SEG, NO_SEG);
             offset++;
@@ -1506,7 +1501,7 @@ static void gencode(int32_t segment, int64_t offset, int bits,
         case 0152:
 	case 0153:
             data = opx->offset;
-            if (is_sbyte32(ins, c & 3)) {
+            if (is_sbyte32(opx)) {
                 bytes[0] = data;
                 out(offset, segment, bytes, OUT_RAWDATA, 1, NO_SEG,
                     NO_SEG);
@@ -1524,7 +1519,7 @@ static void gencode(int32_t segment, int64_t offset, int bits,
 	case 0157:
 	    EMIT_REX();
             bytes[0] = *codes++;
-            if (is_sbyte32(ins, c & 3))
+            if (is_sbyte32(opx))
                 bytes[0] |= 2;  /* s-bit */
             out(offset, segment, bytes, OUT_RAWDATA, 1, NO_SEG, NO_SEG);
             offset++;
@@ -1593,7 +1588,7 @@ static void gencode(int32_t segment, int64_t offset, int bits,
 	case 0253:
             data = opx->offset;
 	    /* is_sbyte32() is right here, we have already warned */
-            if (is_sbyte32(ins, c & 3)) {
+            if (is_sbyte32(opx)) {
                 bytes[0] = data;
                 out(offset, segment, bytes, OUT_RAWDATA, 1, NO_SEG,
                     NO_SEG);
@@ -1718,6 +1713,10 @@ static void gencode(int32_t segment, int64_t offset, int bits,
         case 0335:
 	    break;
 
+        case 0336:
+        case 0337:
+	    break;
+
         case 0340:
             if (ins->oprs[0].segment != NO_SEG)
                 errfunc(ERR_PANIC, "non-constant BSS size in pass two");
@@ -1788,7 +1787,7 @@ static void gencode(int32_t segment, int64_t offset, int bits,
 
                 if (!process_ea
                     (&ins->oprs[(c >> 3) & 7], &ea_data, bits,
-		     ins->addr_size, rfield, rflags, ins->forw_ref)) {
+		     ins->addr_size, rfield, rflags)) {
                     errfunc(ERR_NONFATAL, "invalid effective address");
                 }
 
@@ -2074,8 +2073,10 @@ static int matches(const struct itemplate *itemp, insn * instruction, int bits)
 }
 
 static ea *process_ea(operand * input, ea * output, int bits,
-		      int addrbits, int rfield, int32_t rflags, int forw_ref)
+		      int addrbits, int rfield, int32_t rflags)
 {
+    bool forw_ref = !!(input->opflags & OPFLAG_FORWARD);
+
     output->rip = false;
 
     /* REX flags for the rfield operand */
