@@ -319,8 +319,8 @@ static uint8_t *do_ea(uint8_t *data, int modrm, int asize,
 
             op->scale = 1 << scale;
 
-	    if (index == 4)
-		op->indexreg = -1; /* ESP/RSP/R12 cannot be an index */
+	    if (index == 4 && !(rex & REX_X))
+		op->indexreg = -1; /* ESP/RSP cannot be an index */
 	    else if (a64)
 		op->indexreg = nasm_rd_reg64[index | ((rex & REX_X) ? 8 : 0)];
 	    else
@@ -882,6 +882,10 @@ static int matches(const struct itemplate *t, uint8_t *data,
                 drep = P_REPE;
 	    break;
 
+	case 0336:
+	case 0337:
+	    break;
+
 	case 0340:
 	    return false;
 
@@ -1073,24 +1077,25 @@ int32_t disasm(uint8_t *data, char *output, int outbufsize, int segsize,
 	    if (segsize == 64 || (data[1] & 0xc0) == 0xc0) {
 		prefix.vex[0] = *data++;
 		prefix.vex[1] = *data++;
-		if (prefix.vex[0] == 0xc4)
-		    prefix.vex[2] = *data++;
-	    }
-	    prefix.rex = REX_V;
-	    if (prefix.vex[0] == 0xc4) {
-		prefix.rex |= (~prefix.vex[1] >> 5) & 7; /* REX_RXB */
-		prefix.rex |= (prefix.vex[2] >> (7-3)) & REX_W;
-		prefix.vex_m = prefix.vex[1] & 0x1f;
-		prefix.vex_v = (~prefix.vex[2] >> 3) & 15;
-		prefix.vex_lp = prefix.vex[2] & 7;
-	    } else {
-		prefix.rex |= (~prefix.vex[1] >> (7-2)) & REX_R;
-		prefix.vex_m = 1;
-		prefix.vex_v = (~prefix.vex[1] >> 3) & 15;
-		prefix.vex_lp = prefix.vex[1] & 7;
-	    }
 
-	    ix = itable_VEX[prefix.vex_m][prefix.vex_lp];
+		prefix.rex = REX_V;
+
+		if (prefix.vex[0] == 0xc4) {
+		    prefix.vex[2] = *data++;
+		    prefix.rex |= (~prefix.vex[1] >> 5) & 7; /* REX_RXB */
+		    prefix.rex |= (prefix.vex[2] >> (7-3)) & REX_W;
+		    prefix.vex_m = prefix.vex[1] & 0x1f;
+		    prefix.vex_v = (~prefix.vex[2] >> 3) & 15;
+		    prefix.vex_lp = prefix.vex[2] & 7;
+		} else {
+		    prefix.rex |= (~prefix.vex[1] >> (7-2)) & REX_R;
+		    prefix.vex_m = 1;
+		    prefix.vex_v = (~prefix.vex[1] >> 3) & 15;
+		    prefix.vex_lp = prefix.vex[1] & 7;
+		}
+
+		ix = itable_VEX[prefix.vex_m][prefix.vex_lp];
+	    }
 	    end_prefix = true;
 	    break;
 
