@@ -19,6 +19,8 @@
 
 #include "nasm.h"
 #include "nasmlib.h"
+#include "saa.h"
+#include "raa.h"
 #include "float.h"
 #include "stdscan.h"
 #include "insns.h"
@@ -71,7 +73,9 @@ int optimizing = -1;            /* number of optimization passes to take */
 static int sb, cmd_sb = 16;     /* by default */
 static uint32_t cmd_cpu = IF_PLEVEL;       /* highest level by default */
 static uint32_t cpu = IF_PLEVEL;   /* passed to insn_size & assemble.c */
-bool global_offset_changed;      /* referenced in labels.c */
+int64_t global_offset_changed;      /* referenced in labels.c */
+int64_t prev_offset_changed;
+int32_t stall_count;
 
 static struct location location;
 int in_abs_seg;                 /* Flag we are in ABSOLUTE seg */
@@ -100,18 +104,19 @@ static const char *depend_file = NULL;
  * Which of the suppressible warnings are suppressed. Entry zero
  * isn't an actual warning, but it used for -w+error/-Werror.
  */
-static bool suppressed[ERR_WARN_MAX+1] = {
-    true, false, true, false, false, true, false, true, true, false
-};
+static bool suppressed[ERR_WARN_MAX+1];
 
+static bool suppressed_global[ERR_WARN_MAX+1] = {
+    true, false, true, false, false, false, true, false, true, true, false
+};
 /*
  * The option names for the suppressible warnings. As before, entry
  * zero does nothing.
  */
 static const char *suppressed_names[ERR_WARN_MAX+1] = {
-    "error", "macro-params", "macro-selfref", "orphan-labels",
-    "number-overflow", "gnu-elf-extensions", "float-overflow",
-    "float-denorm", "float-underflow", "float-toolong"
+    "error", "macro-params", "macro-selfref", "macro-defaults",
+    "orphan-labels", "number-overflow", "gnu-elf-extensions",
+    "float-overflow", "float-denorm", "float-underflow", "float-toolong"
 };
 
 /*
@@ -122,6 +127,7 @@ static const char *suppressed_what[ERR_WARN_MAX+1] = {
     "treat warnings as errors",
     "macro calls with wrong parameter count",
     "cyclic macro references",
+    "macros with more default than optional parameters",
     "labels alone on lines without trailing `:'",
     "numeric constants does not fit in 64 bits",
     "using 8- or 16-bit relocation in ELF32, a GNU extension",
@@ -290,12 +296,14 @@ int main(int argc, char **argv)
 
     time(&official_compile_time);
 
-    pass0 = 1;
+    pass0 = 0;
     want_usage = terminate_after_phase = false;
     report_error = report_error_gnu;
 
     error_file = stderr;
 
+    tolower_init();
+
     nasm_set_malloc_error(report_error);
     offsets = raa_init();
     forwrefs = saa_init((int32_t)sizeof(struct forwrefinfo));
@@ -374,7 +382,7 @@ int main(int argc, char **argv)
             location.known = false;
 
 	    /* pass = 1; */
-            preproc->reset(inname, 2, report_error, evaluate, &nasmlist,
+            preproc->reset(inname, 3, report_error, evaluate, &nasmlist,
 			   depend_ptr);
 
             while ((line = preproc->getline())) {
@@ -439,11 +447,12 @@ int main(int argc, char **argv)
                 cleanup_labels();
             } else {
                 /*
-                 * We had an fclose on the output file here, but we
-                 * actually do that in all the object file drivers as well,
-                 * so we're leaving out the one here.
-                 *     fclose (ofile);
+                 * Despite earlier comments, we need this fclose.
+                 * The object output drivers only fclose on cleanup,
+                 * and we just skipped that.
                  */
+                fclose (ofile);
+
                 remove(outname);
                 if (listname[0])
                     remove(listname);
@@ -478,7 +487,7 @@ static char *get_param(char *p, char *q, bool *advance)
     *advance = false;
     if (p[2]) {                 /* the parameter's in the option */
         p += 2;
-        while (isspace(*p))
+        while (nasm_isspace(*p))
             p++;
         return p;
     }
@@ -718,6 +727,7 @@ static bool process_arg(char *p, char *q)
 			     " output format `%s'",
 			     param, ofmt->shortname);
 	    }
+	    using_debug_info = true;
 	    break;
 
 	case 'X':		/* specify error reporting format */
@@ -764,7 +774,7 @@ static bool process_arg(char *p, char *q)
             for (i = 0; i <= ERR_WARN_MAX; i++)
                 printf("    %-23s %s (default %s)\n",
                        suppressed_names[i], suppressed_what[i],
-                       suppressed[i] ? "off" : "on");
+                       suppressed_global[i] ? "off" : "on");
             printf
                 ("\nresponse files should contain command line parameters"
                  ", one per line.\n");
@@ -835,13 +845,13 @@ static bool process_arg(char *p, char *q)
 		if (!nasm_stricmp(param, suppressed_names[i]))
 		    break;
 	    if (i <= ERR_WARN_MAX)
-		suppressed[i] = suppress;
+		suppressed_global[i] = suppress;
 	    else if (!nasm_stricmp(param, "all"))
 		for (i = 1; i <= ERR_WARN_MAX; i++)
-		    suppressed[i] = suppress;
+		    suppressed_global[i] = suppress;
 	    else if (!nasm_stricmp(param, "none"))
 		for (i = 1; i <= ERR_WARN_MAX; i++)
-		    suppressed[i] = !suppress;
+		    suppressed_global[i] = !suppress;
 	    else
 		report_error(ERR_NONFATAL | ERR_NOFILE | ERR_USAGE,
 			     "invalid warning `%s'", param);
@@ -996,11 +1006,11 @@ static void process_respfile(FILE * rfile)
          */
         *(p = &buffer[strcspn(buffer, "\r\n\032")]) = '\0';
 
-        while (p > buffer && isspace(p[-1]))
+        while (p > buffer && nasm_isspace(p[-1]))
             *--p = '\0';
 
         p = buffer;
-        while (isspace(*p))
+        while (nasm_isspace(*p))
             p++;
 
         if (process_arg(prevarg, p))
@@ -1158,8 +1168,7 @@ static void assemble_file(char *fname, StrList **depend_ptr)
         report_error(ERR_FATAL, "command line: "
                      "32-bit segment size requires a higher cpu");
 
-    pass_max = (optimizing > 0 ? optimizing : 0) + 2;   /* passes 1, optimizing, then 2 */
-    pass0 = !(optimizing > 0);  /* start at 1 if not optimizing */
+    pass_max = prev_offset_changed = (INT_MAX >> 1) + 2; /* Almost unlimited */
     for (passn = 1; pass0 <= 2; passn++) {
         int pass1, pass2;
         ldfunc def_label;
@@ -1177,7 +1186,7 @@ static void assemble_file(char *fname, StrList **depend_ptr)
                 nasmlist.init(listname, report_error);
         }
         in_abs_seg = false;
-        global_offset_changed = false;  /* set by redefine_label */
+        global_offset_changed = 0;  /* set by redefine_label */
         location.segment = ofmt->section(NULL, pass2, &sb);
         globalbits = sb;
         if (passn > 1) {
@@ -1188,7 +1197,8 @@ static void assemble_file(char *fname, StrList **depend_ptr)
         }
         preproc->reset(fname, pass1, report_error, evaluate, &nasmlist,
 		       pass1 == 2 ? depend_ptr : NULL);
-	    
+        memcpy(suppressed, suppressed_global, (ERR_WARN_MAX+1) * sizeof(bool));
+
         globallineno = 0;
         if (passn == 1)
             location.known = true;
@@ -1305,7 +1315,7 @@ static void assemble_file(char *fname, StrList **depend_ptr)
                         validid = true;
                         if (!isidstart(*p))
                             validid = false;
-                        while (*p && !isspace(*p)) {
+                        while (*p && !nasm_isspace(*p)) {
                             if (!isidchar(*p))
                                 validid = false;
                             p++;
@@ -1318,7 +1328,7 @@ static void assemble_file(char *fname, StrList **depend_ptr)
                         if (*p) {
                             int64_t size;
 
-                            while (*p && isspace(*p))
+                            while (*p && nasm_isspace(*p))
                                 *p++ = '\0';
                             q = p;
                             while (*q && *q != ':')
@@ -1343,7 +1353,7 @@ static void assemble_file(char *fname, StrList **depend_ptr)
                     } else if (pass0 == 2) {    /* pass == 2 */
                         q = value;
                         while (*q && *q != ':') {
-                            if (isspace(*q))
+                            if (nasm_isspace(*q))
                                 *q = '\0';
                             q++;
                         }
@@ -1383,7 +1393,7 @@ static void assemble_file(char *fname, StrList **depend_ptr)
                     validid = true;
                     if (!isidstart(*p))
                         validid = false;
-                    while (*p && !isspace(*p)) {
+                    while (*p && !nasm_isspace(*p)) {
                         if (!isidchar(*p))
                             validid = false;
                         *q++ = *p++;
@@ -1394,27 +1404,38 @@ static void assemble_file(char *fname, StrList **depend_ptr)
                                      "identifier expected after DEBUG");
                         break;
                     }
-                    while (*p && isspace(*p))
+                    while (*p && nasm_isspace(*p))
                         p++;
                     if (pass0 == 2)
                         ofmt->current_dfmt->debug_directive(debugid, p);
                     break;
-                case D_WARNING:		/* [WARNING {+|-}warn-name] */
+                case D_WARNING:		/* [WARNING {+|-|*}warn-name] */
                     if (pass1 == 1) {
-                        while (*value && isspace(*value))
+                        while (*value && nasm_isspace(*value))
                             value++;
 
-                        if (*value == '+' || *value == '-') {
-                            validid = (*value == '-') ? true : false;
-                            value++;
-                        } else
-                            validid = false;
+                        switch(*value) {
+                            case '-': validid = 0; value++; break;
+                            case '+': validid = 1; value++; break;
+                            case '*': validid = 2; value++; break;
+                            default: /*
+                                       * Should this error out?
+                                       * I'll keep it so nothing breaks.
+                                       */
+                                      validid = 1; break;
+                        }
 
                         for (i = 1; i <= ERR_WARN_MAX; i++)
                             if (!nasm_stricmp(value, suppressed_names[i]))
                                 break;
-                        if (i <= ERR_WARN_MAX)
-                            suppressed[i] = validid;
+                        if (i <= ERR_WARN_MAX) {
+                            switch(validid) {
+                                case 0: suppressed[i] = true; break;
+                                case 1: suppressed[i] = false; break;
+                                case 2: suppressed[i] = suppressed_global[i];
+                                        break;
+                            }
+                        }
                         else
                             report_error(ERR_NONFATAL,
                                          "invalid warning id in WARNING directive");
@@ -1424,7 +1445,7 @@ static void assemble_file(char *fname, StrList **depend_ptr)
                     cpu = get_cpu(value);
                     break;
                 case D_LIST:		/* [LIST {+|-}] */
-                    while (*value && isspace(*value))
+                    while (*value && nasm_isspace(*value))
                         value++;
 
                     if (*value == '+') {
@@ -1480,7 +1501,7 @@ static void assemble_file(char *fname, StrList **depend_ptr)
                 parse_line(pass1, line, &output_ins,
                            report_error, evaluate, def_label);
 
-                if (!(optimizing > 0) && pass0 == 2) {
+                if (optimizing > 0) {
                     if (forwref != NULL && globallineno == forwref->lineno) {
                         output_ins.forw_ref = true;
                         do {
@@ -1493,7 +1514,7 @@ static void assemble_file(char *fname, StrList **depend_ptr)
                         output_ins.forw_ref = false;
                 }
 
-                if (!(optimizing > 0) && output_ins.forw_ref) {
+                if (optimizing > 0) {
                     if (passn == 1) {
                         for (i = 0; i < output_ins.operands; i++) {
                             if (output_ins.oprs[i].
@@ -1505,36 +1526,7 @@ static void assemble_file(char *fname, StrList **depend_ptr)
                                 fwinf->operand = i;
                             }
                         }
-                    } else {    /* passn > 1 */
-                        /*
-                         * Hack to prevent phase error in the code
-                         *   rol ax,x
-                         *   x equ 1
-                         *
-                         * If the second operand is a forward reference,
-                         * the UNITY property of the number 1 in that
-                         * operand is cancelled. Otherwise the above
-                         * sequence will cause a phase error.
-                         *
-                         * This hack means that the above code will
-                         * generate 286+ code.
-                         *
-                         * The forward reference will mean that the
-                         * operand will not have the UNITY property on
-                         * the first pass, so the pass behaviours will
-                         * be consistent.
-                         */
-
-                        if (output_ins.operands >= 2 &&
-                            (output_ins.oprs[1].opflags & OPFLAG_FORWARD) &&
-			    !(IMMEDIATE & ~output_ins.oprs[1].type))
-			{
-			    /* Remove special properties bits */
-			    output_ins.oprs[1].type &= ~REG_SMASK;
-                        }
-
                     }
-
                 }
 
                 /*  forw_ref */
@@ -1725,7 +1717,6 @@ static void assemble_file(char *fname, StrList **depend_ptr)
             nasm_free(line);
             location.offset = offs = GET_CURR_OFFS;
         }                       /* end while (line = preproc->getline... */
-
         if (pass1 == 2 && global_offset_changed)
             report_error(ERR_NONFATAL,
                          "phase error detected at end of assembly.");
@@ -1740,18 +1731,30 @@ static void assemble_file(char *fname, StrList **depend_ptr)
                 usage();
             exit(1);
         }
-        if (passn >= pass_max - 2 ||
-	    (passn > 1 && !global_offset_changed))
+
+        if (passn > 1 && !global_offset_changed)
             pass0++;
+        else if (global_offset_changed && global_offset_changed < prev_offset_changed) {
+            prev_offset_changed = global_offset_changed;
+            stall_count = 0;
+            }
+        else stall_count++;
+
+        if((stall_count > 997) || (passn >= pass_max))
+            /* We get here if the labels don't converge
+             * Example: FOO equ FOO + 1
+             */
+             report_error(ERR_NONFATAL,
+                          "Can't find valid values for all labels "
+                          "after %d passes, giving up.\n"
+                          "                  Possible cause: recursive equ's.", passn);
     }
 
     preproc->cleanup(0);
     nasmlist.cleanup();
-#if 1
-    if (optimizing > 0 && opt_verbose_info)     /*  -On and -Ov switches */
+    if (opt_verbose_info)     /*  -On and -Ov switches */
         fprintf(stdout,
                 "info:: assembly required 1+%d+1 passes\n", passn-3);
-#endif
 }                               /* exit from assemble_file (...) */
 
 static enum directives getkw(char **directive, char **value)
@@ -1778,7 +1781,7 @@ static enum directives getkw(char **directive, char **value)
     q = p++;
 
     while (*p && *p != ';') {
-        if (!isspace(*p))
+        if (!nasm_isspace(*p))
             return 0;
         p++;
     }
@@ -1792,7 +1795,7 @@ static enum directives getkw(char **directive, char **value)
         *value = buf;
     } else {
         *buf++ = '\0';
-        while (isspace(*buf))
+        while (nasm_isspace(*buf))
             buf++;              /* beppu - skip leading whitespace */
         *value = buf;
         while (*buf != ']')
@@ -1907,7 +1910,7 @@ static bool is_suppressed_warning(int severity)
 static void report_error_common(int severity, const char *fmt,
                                 va_list args)
 {
-    switch (severity & ERR_MASK) {
+    switch (severity & (ERR_MASK|ERR_NO_SEVERITY)) {
     case ERR_WARNING:
         fputs("warning: ", error_file);
         break;
@@ -1923,6 +1926,8 @@ static void report_error_common(int severity, const char *fmt,
     case ERR_DEBUG:
         fputs("debug: ", error_file);
         break;
+    default:
+	break;
     }
 
     vfprintf(error_file, fmt, args);
