@@ -17,6 +17,7 @@
 #include <inttypes.h>
 #include "version.h"            /* generated NASM version macros */
 #include "nasmlib.h"
+#include "preproc.h"
 #include "insnsi.h"		/* For enum opcode */
 
 #define NO_SEG -1L              /* null segment value */
@@ -165,8 +166,14 @@ enum token_type {		/* token types, other than chars */
     TOKEN_INVALID = -1,         /* a placeholder value */
     TOKEN_EOS = 0,              /* end of string */
     TOKEN_EQ = '=', TOKEN_GT = '>', TOKEN_LT = '<',     /* aliases */
-    TOKEN_ID = 256, TOKEN_NUM, TOKEN_REG, TOKEN_INSN,   /* major token types */
-    TOKEN_ERRNUM,               /* numeric constant with error in */
+    TOKEN_ID = 256,		/* identifier */
+    TOKEN_NUM,			/* numeric constant */
+    TOKEN_ERRNUM,		/* malformed numeric constant */
+    TOKEN_STR,			/* string constant */
+    TOKEN_ERRSTR,               /* unterminated string constant */
+    TOKEN_FLOAT,                /* floating-point constant */
+    TOKEN_REG,			/* register name */
+    TOKEN_INSN,			/* instruction name */
     TOKEN_HERE, TOKEN_BASE,     /* $ and $$ */
     TOKEN_SPECIAL,              /* BYTE, WORD, DWORD, QWORD, FAR, NEAR, etc */
     TOKEN_PREFIX,               /* A32, O16, LOCK, REPNZ, TIMES, etc */
@@ -175,8 +182,8 @@ enum token_type {		/* token types, other than chars */
     TOKEN_GE, TOKEN_LE, TOKEN_NE,       /* >=, <= and <> (!= is same as <>) */
     TOKEN_DBL_AND, TOKEN_DBL_OR, TOKEN_DBL_XOR, /* &&, || and ^^ */
     TOKEN_SEG, TOKEN_WRT,       /* SEG and WRT */
-    TOKEN_FLOAT,                /* floating-point constant */
     TOKEN_FLOATIZE,		/* __floatX__ */
+    TOKEN_STRFUNC,		/* __utf16__, __utf32__ */
 };
 
 enum floatize {
@@ -190,6 +197,14 @@ enum floatize {
     FLOAT_128H,
 };
 
+/* Must match the list in string_transform(), in strfunc.c */
+enum strfunc {
+    STRFUNC_UTF16,
+    STRFUNC_UTF32,
+};
+
+size_t string_transform(char *, size_t, char **, enum strfunc);
+
 /*
  * The expression evaluator must be passed a scanner function; a
  * standard scanner is provided as part of nasmlib.c. The
@@ -339,15 +354,15 @@ extern Preproc nasmpp;
  * start.
  */
 
-#define isidstart(c) ( isalpha(c) || (c)=='_' || (c)=='.' || (c)=='?' \
+#define isidstart(c) ( nasm_isalpha(c) || (c)=='_' || (c)=='.' || (c)=='?' \
                                   || (c)=='@' )
-#define isidchar(c)  ( isidstart(c) || isdigit(c) || (c)=='$' || (c)=='#' \
-                                                  || (c)=='~' )
+#define isidchar(c)  ( isidstart(c) || nasm_isdigit(c) || \
+		       (c)=='$' || (c)=='#' || (c)=='~' )
 
 /* Ditto for numeric constants. */
 
-#define isnumstart(c)  ( isdigit(c) || (c)=='$' )
-#define isnumchar(c)   ( isalnum(c) || (c)=='_' )
+#define isnumstart(c)  ( nasm_isdigit(c) || (c)=='$' )
+#define isnumchar(c)   ( nasm_isalnum(c) || (c)=='_' )
 
 /* This returns the numeric value of a given 'digit'. */
 
@@ -600,11 +615,14 @@ enum prefixes {			/* instruction prefixes */
     PREFIX_ENUM_LIMIT
 };
 
-enum {                          /* extended operand types */
-    EOT_NOTHING, EOT_DB_STRING, EOT_DB_NUMBER
+enum extop_type {		/* extended operand types */
+    EOT_NOTHING,
+    EOT_DB_STRING,		/* Byte string */
+    EOT_DB_STRING_FREE,		/* Byte string which should be nasm_free'd*/
+    EOT_DB_NUMBER,		/* Integer */
 };
 
-enum {                          /* special EA flags */
+enum ea_flags {			/* special EA flags */
     EAF_BYTEOFFS =  1,          /* force offset part to byte size */
     EAF_WORDOFFS =  2,          /* force offset part to [d]word size */
     EAF_TIMESTWO =  4,          /* really do EAX*2 not EAX+EAX */
@@ -638,12 +656,12 @@ typedef struct operand {	/* operand to an instruction */
 
 typedef struct extop {          /* extended operand */
     struct extop *next;         /* linked list */
-    int32_t type;               /* defined above */
-    char *stringval;          /* if it's a string, then here it is */
-    int stringlen;              /* ... and here's how long it is */
-    int32_t segment;            /* if it's a number/address, then... */
+    char *stringval;	        /* if it's a string, then here it is */
+    size_t stringlen;           /* ... and here's how long it is */
     int64_t offset;             /* ... it's given here ... */
+    int32_t segment;            /* if it's a number/address, then... */
     int32_t wrt;                /* ... and here */
+    enum extop_type type;	/* defined above */
 } extop;
 
 /* Prefix positions: each type of prefix goes in a specific slot.
@@ -677,7 +695,7 @@ typedef struct insn {		/* an instruction itself */
     extop *eops;                /* extended operands */
     int eops_float;             /* true if DD and floating */
     int32_t times;              /* repeat count (TIMES prefix) */
-    int forw_ref;               /* is there a forward reference? */
+    bool forw_ref;              /* is there a forward reference? */
     int rex;			/* Special REX Prefix */
     int drexdst;		/* Destination register for DREX/VEX suffix */
     int vex_m;			/* M register for VEX prefix */
@@ -733,7 +751,7 @@ struct ofmt {
      * and user-level equivalents for any format-specific
      * directives).
      */
-    const char **stdmac;
+    macros_t *stdmac;
 
     /*
      * This procedure is called at the start of an output session.
