@@ -54,6 +54,7 @@
 
 #include "nasm.h"
 #include "nasmlib.h"
+#include "saa.h"
 #include "stdscan.h"
 #include "labels.h"
 #include "eval.h"
@@ -138,15 +139,7 @@ static int origin_defined;
 static int map_control = 0;
 static char *infile, *outfile;
 
-static const char *bin_stdmac[] = {
-    "%define __SECT__ [section .text]",
-    "%imacro org 1+.nolist",
-    "[org %1]",
-    "%endmacro",
-    "%macro __NASM_CDecl__ 1",
-    "%endmacro",
-    NULL
-};
+extern macros_t bin_stdmac[];
 
 static void add_reloc(struct Section *s, int32_t bytes, int32_t secref,
                       int32_t secrel)
@@ -888,7 +881,7 @@ static int bin_read_attribute(char **line, int *attribute,
     char *exp;
 
     /* Skip whitespace. */
-    while (**line && isspace(**line))
+    while (**line && nasm_isspace(**line))
         (*line)++;
     if (!**line)
         return 0;
@@ -916,12 +909,12 @@ static int bin_read_attribute(char **line, int *attribute,
             *line += 9;
             return 1;
         } else if (!nasm_strnicmp(*line, "nobits", 6) &&
-                   (isspace((*line)[6]) || ((*line)[6] == '\0'))) {
+                   (nasm_isspace((*line)[6]) || ((*line)[6] == '\0'))) {
             *attribute = ATTRIB_NOBITS;
             *line += 6;
             return 1;
         } else if (!nasm_strnicmp(*line, "progbits", 8) &&
-                   (isspace((*line)[8]) || ((*line)[8] == '\0'))) {
+                   (nasm_isspace((*line)[8]) || ((*line)[8] == '\0'))) {
             *attribute = ATTRIB_PROGBITS;
             *line += 8;
             return 1;
@@ -934,7 +927,7 @@ static int bin_read_attribute(char **line, int *attribute,
     if ((*line)[attrib_name_size] != '(') {
         /* Single term (no parenthesis). */
         exp = *line += attrib_name_size;
-        while (**line && !isspace(**line))
+        while (**line && !nasm_isspace(**line))
             (*line)++;
         if (**line) {
             **line = '\0';
@@ -1024,7 +1017,7 @@ static void bin_assign_attributes(struct Section *sec, char *astring)
                 break;          /* End of line. */
             else {
                 p = astring;
-                while (*astring && !isspace(*astring))
+                while (*astring && !nasm_isspace(*astring))
                     astring++;
                 if (*astring) {
                     *astring = '\0';
@@ -1252,7 +1245,7 @@ static int32_t bin_secname(char *name, int pass, int *bits)
     /* Attempt to find the requested section.  If it does not
      * exist, create it. */
     p = name;
-    while (*p && !isspace(*p))
+    while (*p && !nasm_isspace(*p))
         p++;
     if (*p)
         *p++ = '\0';
