@@ -16,9 +16,10 @@
 
 #include "nasm.h"
 #include "nasmlib.h"
+#include "saa.h"
+#include "raa.h"
 #include "stdscan.h"
 #include "outform.h"
-#include "wsaa.h"
 
 /* Definitions in lieu of elf.h */
 #define SHT_NULL 0			/* Inactive section header */
@@ -267,7 +268,6 @@ static int8_t line_base = -5, line_range = 14, opcode_base = 13;
 static int arangeslen, arangesrellen, pubnameslen, infolen, inforellen,
            abbrevlen, linelen, linerellen, framelen, loclen;
 static int64_t dwarf_infosym, dwarf_abbrevsym, dwarf_linesym;
-static char workbuf[1024];
 
 
 static struct dfmt df_dwarf;
@@ -432,21 +432,21 @@ static int32_t elf_section_names(char *name, int pass, int *bits)
     }
 
     p = name;
-    while (*p && !isspace(*p))
+    while (*p && !nasm_isspace(*p))
         p++;
     if (*p)
         *p++ = '\0';
     flags_and = flags_or = type = align = 0;
 
-    while (*p && isspace(*p))
+    while (*p && nasm_isspace(*p))
         p++;
     while (*p) {
         char *q = p;
-        while (*p && !isspace(*p))
+        while (*p && !nasm_isspace(*p))
             p++;
         if (*p)
             *p++ = '\0';
-        while (*p && isspace(*p))
+        while (*p && nasm_isspace(*p))
             p++;
 
         if (!nasm_strnicmp(q, "align=", 6)) {
@@ -562,9 +562,9 @@ static void elf_deflabel(char *name, int32_t segment, int64_t offset,
                 expr *e;
                 char *p = special;
 
-                while (*p && !isspace(*p))
+                while (*p && !nasm_isspace(*p))
                     p++;
-                while (*p && isspace(*p))
+                while (*p && nasm_isspace(*p))
                     p++;
                 stdscan_reset();
                 stdscan_bufptr = p;
@@ -681,7 +681,7 @@ static void elf_deflabel(char *name, int32_t segment, int64_t offset,
                           n, special);
                 special += n;
 
-                while (isspace(*special))
+                while (nasm_isspace(*special))
                     ++special;
                 if (*special) {
                     n = strcspn(special, " \t");
@@ -704,7 +704,7 @@ static void elf_deflabel(char *name, int32_t segment, int64_t offset,
 		    int fwd = 0;
                     char *saveme = stdscan_bufptr;      /* bugfix? fbk 8/10/00 */
 
-                    while (special[n] && isspace(special[n]))
+                    while (special[n] && nasm_isspace(special[n]))
                         n++;
                     /*
                      * We have a size expression; attempt to
@@ -1511,16 +1511,8 @@ static void elf_filename(char *inname, char *outname, efunc error)
     standard_extension(inname, outname, ".o", error);
 }
 
-static const char *elf_stdmac[] = {
-    "%define __SECT__ [section .text]",
-    "%macro __NASM_CDecl__ 1",
-    "%define $_%1 $%1",
-    "%endmacro",
-    "%macro osabi 1+.nolist",
-    "[osabi %1]",
-    "%endmacro",
-    NULL
-};
+extern macros_t elf_stdmac[];
+
 static int elf_set_info(enum geninfo type, char **val)
 {
     (void)type;
@@ -1915,8 +1907,8 @@ void dwarf64_output(int type, void *param)
     /* check for file change */
     if (!(inx == dwarf_csect->file))
     {
-       WSAACHAR(plinep,workbuf,DW_LNS_set_file);
-       WSAACHAR(plinep,workbuf,inx);
+       saa_write8(plinep,DW_LNS_set_file);
+       saa_write8(plinep,inx);
        dwarf_csect->file = inx;
     }
     /* check for line change */
@@ -1927,18 +1919,18 @@ void dwarf64_output(int type, void *param)
        soc = (ln - line_base) + (line_range * aa) + opcode_base;
        if (ln >= line_base && ln < maxln && soc < 256)
        {
-          WSAACHAR(plinep,workbuf,soc);
+          saa_write8(plinep,soc);
        }
        else
        {
           if (ln)
           {
-          WSAACHAR(plinep,workbuf,DW_LNS_advance_line);
+          saa_write8(plinep,DW_LNS_advance_line);
           saa_wleb128s(plinep,ln);
           }
           if (aa)
           {
-          WSAACHAR(plinep,workbuf,DW_LNS_advance_pc);
+          saa_write8(plinep,DW_LNS_advance_pc);
           saa_wleb128u(plinep,aa);
           }
        }
@@ -1953,6 +1945,7 @@ void dwarf64_output(int type, void *param)
 
 void dwarf64_generate(void)
 {
+    static const char nasm_signature[] = "NASM " NASM_VER;
     uint8_t *pbuf;
     int indx;
     struct linelist *ftentry;
@@ -1965,14 +1958,14 @@ void dwarf64_generate(void)
     /* and build aranges section */
     paranges = saa_init(1L);
     parangesrel = saa_init(1L);
-    WSAASHORT(paranges,workbuf,3);		/* dwarf version */
-    WSAADLONG(parangesrel,workbuf, paranges->datalen+4);
-    WSAADLONG(parangesrel,workbuf, (dwarf_infosym << 32) +  R_X86_64_32); /* reloc to info */
-    WSAADLONG(parangesrel,workbuf, (uint64_t) 0);
-    WSAALONG(paranges,workbuf,0);		/* offset into info */
-    WSAACHAR(paranges,workbuf,8);		/* pointer size */
-    WSAACHAR(paranges,workbuf,0);		/* not segmented */
-    WSAALONG(paranges,workbuf,0);		/* padding */
+    saa_write16(paranges,3);		/* dwarf version */
+    saa_write64(parangesrel, paranges->datalen+4);
+    saa_write64(parangesrel, (dwarf_infosym << 32) +  R_X86_64_32); /* reloc to info */
+    saa_write64(parangesrel, 0);
+    saa_write32(paranges,0);		/* offset into info */
+    saa_write8(paranges,8);		/* pointer size */
+    saa_write8(paranges,0);		/* not segmented */
+    saa_write32(paranges,0);		/* padding */
     /* iterate though sectlist entries */
      psect = dwarf_fsect;
      totlen = 0;
@@ -1981,25 +1974,25 @@ void dwarf64_generate(void)
      {
          plinep = psect->psaa;
          /* Line Number Program Epilogue */
-         WSAACHAR(plinep,workbuf,2);			/* std op 2 */
-         WSAACHAR(plinep,workbuf,(sects[psect->section]->len)-psect->offset);
-         WSAACHAR(plinep,workbuf,DW_LNS_extended_op);
-         WSAACHAR(plinep,workbuf,1);			/* operand length */
-         WSAACHAR(plinep,workbuf,DW_LNE_end_sequence);
+         saa_write8(plinep,2);			/* std op 2 */
+         saa_write8(plinep,(sects[psect->section]->len)-psect->offset);
+         saa_write8(plinep,DW_LNS_extended_op);
+         saa_write8(plinep,1);			/* operand length */
+         saa_write8(plinep,DW_LNE_end_sequence);
          totlen += plinep->datalen;
          /* range table relocation entry */
-         WSAADLONG(parangesrel,workbuf, paranges->datalen + 4);
-         WSAADLONG(parangesrel,workbuf, ((uint64_t) (psect->section + 2) << 32) +  R_X86_64_64);
-         WSAADLONG(parangesrel,workbuf, (uint64_t) 0);
+         saa_write64(parangesrel, paranges->datalen + 4);
+         saa_write64(parangesrel, ((uint64_t) (psect->section + 2) << 32) +  R_X86_64_64);
+         saa_write64(parangesrel, (uint64_t) 0);
          /* range table entry */
-         WSAADLONG(paranges,workbuf,0x0000);		/* range start */
-         WSAADLONG(paranges,workbuf,sects[psect->section]->len);	/* range length */
+         saa_write64(paranges,0x0000);		/* range start */
+         saa_write64(paranges,sects[psect->section]->len);	/* range length */
          highaddr += sects[psect->section]->len;
          /* done with this entry */
          psect = psect->next;
      }
-    WSAADLONG(paranges,workbuf,0);		/* null address */
-    WSAADLONG(paranges,workbuf,0);		/* null length */
+    saa_write64(paranges,0);		/* null address */
+    saa_write64(paranges,0);		/* null length */
     saalen = paranges->datalen;
     arangeslen = saalen + 4;
     arangesbuf = pbuf = nasm_malloc(arangeslen);
@@ -2015,10 +2008,10 @@ void dwarf64_generate(void)
 
     /* build pubnames section */
     ppubnames = saa_init(1L);
-    WSAASHORT(ppubnames,workbuf,3);			/* dwarf version */
-    WSAALONG(ppubnames,workbuf,0);			/* offset into info */
-    WSAALONG(ppubnames,workbuf,0);			/* space used in info */
-    WSAALONG(ppubnames,workbuf,0);			/* end of list */
+    saa_write16(ppubnames,3);			/* dwarf version */
+    saa_write32(ppubnames,0);			/* offset into info */
+    saa_write32(ppubnames,0);			/* space used in info */
+    saa_write32(ppubnames,0);			/* end of list */
     saalen = ppubnames->datalen;
     pubnameslen = saalen + 4;
     pubnamesbuf = pbuf = nasm_malloc(pubnameslen);
@@ -2029,37 +2022,35 @@ void dwarf64_generate(void)
     /* build info section */
     pinfo = saa_init(1L);
     pinforel = saa_init(1L);
-    WSAASHORT(pinfo,workbuf,3);			/* dwarf version */
-    WSAADLONG(pinforel,workbuf, pinfo->datalen + 4);
-    WSAADLONG(pinforel,workbuf, (dwarf_abbrevsym << 32) +  R_X86_64_32); /* reloc to abbrev */
-    WSAADLONG(pinforel,workbuf, (uint64_t) 0);
-    WSAALONG(pinfo,workbuf,0);			/* offset into abbrev */
-    WSAACHAR(pinfo,workbuf,8);			/* pointer size */
-    WSAACHAR(pinfo,workbuf,1);			/* abbrviation number LEB128u */
-    WSAADLONG(pinforel,workbuf, pinfo->datalen + 4);
-    WSAADLONG(pinforel,workbuf, ((uint64_t)(dwarf_fsect->section + 2) << 32) +  R_X86_64_64);
-    WSAADLONG(pinforel,workbuf, (uint64_t) 0);
-    WSAADLONG(pinfo,workbuf,0);			/* DW_AT_low_pc */
-    WSAADLONG(pinforel,workbuf, pinfo->datalen + 4);
-    WSAADLONG(pinforel,workbuf, ((uint64_t)(dwarf_fsect->section + 2) << 32) +  R_X86_64_64);
-    WSAADLONG(pinforel,workbuf, (uint64_t) 0);
-    WSAADLONG(pinfo,workbuf,highaddr);		/* DW_AT_high_pc */
-    WSAADLONG(pinforel,workbuf, pinfo->datalen + 4);
-    WSAADLONG(pinforel,workbuf, (dwarf_linesym << 32) +  R_X86_64_32); /* reloc to line */
-    WSAADLONG(pinforel,workbuf, (uint64_t) 0);
-    WSAALONG(pinfo,workbuf,0);			/* DW_AT_stmt_list */
-    strcpy(workbuf,elf_module);    		/* input file name */
-    saa_wbytes(pinfo, workbuf, (int32_t)(strlen(elf_module) + 1));
-    sprintf(workbuf, "NASM %s", NASM_VER);
-    saa_wbytes(pinfo, workbuf, (int32_t)(strlen(workbuf) + 1));
-    WSAASHORT(pinfo,workbuf,DW_LANG_Mips_Assembler);
-    WSAACHAR(pinfo,workbuf,2);			/* abbrviation number LEB128u */
-    WSAADLONG(pinforel,workbuf, pinfo->datalen + 4);
-    WSAADLONG(pinforel,workbuf, ((uint64_t)(dwarf_fsect->section + 2) << 32) +  R_X86_64_64);
-    WSAADLONG(pinforel,workbuf, (uint64_t) 0);
-    WSAADLONG(pinfo,workbuf,0);			/* DW_AT_low_pc */
-    WSAADLONG(pinfo,workbuf,0);			/* DW_AT_frame_base */
-    WSAACHAR(pinfo,workbuf,0);			/* end of entries */
+    saa_write16(pinfo,3);			/* dwarf version */
+    saa_write64(pinforel, pinfo->datalen + 4);
+    saa_write64(pinforel, (dwarf_abbrevsym << 32) +  R_X86_64_32); /* reloc to abbrev */
+    saa_write64(pinforel, 0);
+    saa_write32(pinfo,0);			/* offset into abbrev */
+    saa_write8(pinfo,8);			/* pointer size */
+    saa_write8(pinfo,1);			/* abbrviation number LEB128u */
+    saa_write64(pinforel, pinfo->datalen + 4);
+    saa_write64(pinforel, ((uint64_t)(dwarf_fsect->section + 2) << 32) +  R_X86_64_64);
+    saa_write64(pinforel, 0);
+    saa_write64(pinfo,0);			/* DW_AT_low_pc */
+    saa_write64(pinforel, pinfo->datalen + 4);
+    saa_write64(pinforel, ((uint64_t)(dwarf_fsect->section + 2) << 32) +  R_X86_64_64);
+    saa_write64(pinforel, 0);
+    saa_write64(pinfo,highaddr);		/* DW_AT_high_pc */
+    saa_write64(pinforel, pinfo->datalen + 4);
+    saa_write64(pinforel, (dwarf_linesym << 32) +  R_X86_64_32); /* reloc to line */
+    saa_write64(pinforel, 0);
+    saa_write32(pinfo,0);			/* DW_AT_stmt_list */
+    saa_wbytes(pinfo, elf_module, strlen(elf_module)+1);
+    saa_wbytes(pinfo, nasm_signature, strlen(nasm_signature)+1);
+    saa_write16(pinfo,DW_LANG_Mips_Assembler);
+    saa_write8(pinfo,2);			/* abbrviation number LEB128u */
+    saa_write64(pinforel, pinfo->datalen + 4);
+    saa_write64(pinforel, ((uint64_t)(dwarf_fsect->section + 2) << 32) +  R_X86_64_64);
+    saa_write64(pinforel, 0);
+    saa_write64(pinfo,0);			/* DW_AT_low_pc */
+    saa_write64(pinfo,0);			/* DW_AT_frame_base */
+    saa_write8(pinfo,0);			/* end of entries */
     saalen = pinfo->datalen;
     infolen = saalen + 4;
     infobuf = pbuf = nasm_malloc(infolen);
@@ -2075,32 +2066,32 @@ void dwarf64_generate(void)
 
     /* build abbrev section */
     pabbrev = saa_init(1L);
-    WSAACHAR(pabbrev,workbuf,1);			/* entry number LEB128u */
-    WSAACHAR(pabbrev,workbuf,DW_TAG_compile_unit);	/* tag LEB128u */
-    WSAACHAR(pabbrev,workbuf,1);			/* has children */
+    saa_write8(pabbrev,1);			/* entry number LEB128u */
+    saa_write8(pabbrev,DW_TAG_compile_unit);	/* tag LEB128u */
+    saa_write8(pabbrev,1);			/* has children */
     /* the following attributes and forms are all LEB128u values */
-    WSAACHAR(pabbrev,workbuf,DW_AT_low_pc);
-    WSAACHAR(pabbrev,workbuf,DW_FORM_addr);
-    WSAACHAR(pabbrev,workbuf,DW_AT_high_pc);
-    WSAACHAR(pabbrev,workbuf,DW_FORM_addr);
-    WSAACHAR(pabbrev,workbuf,DW_AT_stmt_list);
-    WSAACHAR(pabbrev,workbuf,DW_FORM_data4);
-    WSAACHAR(pabbrev,workbuf,DW_AT_name);
-    WSAACHAR(pabbrev,workbuf,DW_FORM_string);
-    WSAACHAR(pabbrev,workbuf,DW_AT_producer);
-    WSAACHAR(pabbrev,workbuf,DW_FORM_string);
-    WSAACHAR(pabbrev,workbuf,DW_AT_language);
-    WSAACHAR(pabbrev,workbuf,DW_FORM_data2);
-    WSAASHORT(pabbrev,workbuf,0);			/* end of entry */
+    saa_write8(pabbrev,DW_AT_low_pc);
+    saa_write8(pabbrev,DW_FORM_addr);
+    saa_write8(pabbrev,DW_AT_high_pc);
+    saa_write8(pabbrev,DW_FORM_addr);
+    saa_write8(pabbrev,DW_AT_stmt_list);
+    saa_write8(pabbrev,DW_FORM_data4);
+    saa_write8(pabbrev,DW_AT_name);
+    saa_write8(pabbrev,DW_FORM_string);
+    saa_write8(pabbrev,DW_AT_producer);
+    saa_write8(pabbrev,DW_FORM_string);
+    saa_write8(pabbrev,DW_AT_language);
+    saa_write8(pabbrev,DW_FORM_data2);
+    saa_write16(pabbrev,0);			/* end of entry */
     /* LEB128u usage same as above */
-    WSAACHAR(pabbrev,workbuf,2);			/* entry number */
-    WSAACHAR(pabbrev,workbuf,DW_TAG_subprogram);
-    WSAACHAR(pabbrev,workbuf,0);			/* no children */
-    WSAACHAR(pabbrev,workbuf,DW_AT_low_pc);
-    WSAACHAR(pabbrev,workbuf,DW_FORM_addr);
-    WSAACHAR(pabbrev,workbuf,DW_AT_frame_base);
-    WSAACHAR(pabbrev,workbuf,DW_FORM_data4);
-    WSAASHORT(pabbrev,workbuf,0);			/* end of entry */
+    saa_write8(pabbrev,2);			/* entry number */
+    saa_write8(pabbrev,DW_TAG_subprogram);
+    saa_write8(pabbrev,0);			/* no children */
+    saa_write8(pabbrev,DW_AT_low_pc);
+    saa_write8(pabbrev,DW_FORM_addr);
+    saa_write8(pabbrev,DW_AT_frame_base);
+    saa_write8(pabbrev,DW_FORM_data4);
+    saa_write16(pabbrev,0);			/* end of entry */
     abbrevlen = saalen = pabbrev->datalen;
     abbrevbuf = pbuf = nasm_malloc(saalen);
     saa_rnbytes(pabbrev, pbuf, saalen);
@@ -2109,37 +2100,37 @@ void dwarf64_generate(void)
     /* build line section */
     /* prolog */
     plines = saa_init(1L);
-    WSAACHAR(plines,workbuf,1);			/* Minimum Instruction Length */
-    WSAACHAR(plines,workbuf,1);			/* Initial value of 'is_stmt' */
-    WSAACHAR(plines,workbuf,line_base);		/* Line Base */
-    WSAACHAR(plines,workbuf,line_range);	/* Line Range */
-    WSAACHAR(plines,workbuf,opcode_base);	/* Opcode Base */
+    saa_write8(plines,1);			/* Minimum Instruction Length */
+    saa_write8(plines,1);			/* Initial value of 'is_stmt' */
+    saa_write8(plines,line_base);		/* Line Base */
+    saa_write8(plines,line_range);	/* Line Range */
+    saa_write8(plines,opcode_base);	/* Opcode Base */
     /* standard opcode lengths (# of LEB128u operands) */
-    WSAACHAR(plines,workbuf,0);			/* Std opcode 1 length */
-    WSAACHAR(plines,workbuf,1);			/* Std opcode 2 length */
-    WSAACHAR(plines,workbuf,1);			/* Std opcode 3 length */
-    WSAACHAR(plines,workbuf,1);			/* Std opcode 4 length */
-    WSAACHAR(plines,workbuf,1);			/* Std opcode 5 length */
-    WSAACHAR(plines,workbuf,0);			/* Std opcode 6 length */
-    WSAACHAR(plines,workbuf,0);			/* Std opcode 7 length */
-    WSAACHAR(plines,workbuf,0);			/* Std opcode 8 length */
-    WSAACHAR(plines,workbuf,1);			/* Std opcode 9 length */
-    WSAACHAR(plines,workbuf,0);			/* Std opcode 10 length */
-    WSAACHAR(plines,workbuf,0);			/* Std opcode 11 length */
-    WSAACHAR(plines,workbuf,1);			/* Std opcode 12 length */
+    saa_write8(plines,0);			/* Std opcode 1 length */
+    saa_write8(plines,1);			/* Std opcode 2 length */
+    saa_write8(plines,1);			/* Std opcode 3 length */
+    saa_write8(plines,1);			/* Std opcode 4 length */
+    saa_write8(plines,1);			/* Std opcode 5 length */
+    saa_write8(plines,0);			/* Std opcode 6 length */
+    saa_write8(plines,0);			/* Std opcode 7 length */
+    saa_write8(plines,0);			/* Std opcode 8 length */
+    saa_write8(plines,1);			/* Std opcode 9 length */
+    saa_write8(plines,0);			/* Std opcode 10 length */
+    saa_write8(plines,0);			/* Std opcode 11 length */
+    saa_write8(plines,1);			/* Std opcode 12 length */
     /* Directory Table */ 
-    WSAACHAR(plines,workbuf,0);			/* End of table */
+    saa_write8(plines,0);			/* End of table */
     /* File Name Table */
     ftentry = dwarf_flist;
     for (indx = 0;indx<dwarf_numfiles;indx++)
     {
       saa_wbytes(plines, ftentry->filename, (int32_t)(strlen(ftentry->filename) + 1));
-      WSAACHAR(plines,workbuf,0);			/* directory  LEB128u */
-      WSAACHAR(plines,workbuf,0);			/* time LEB128u */
-      WSAACHAR(plines,workbuf,0);			/* size LEB128u */
+      saa_write8(plines,0);			/* directory  LEB128u */
+      saa_write8(plines,0);			/* time LEB128u */
+      saa_write8(plines,0);			/* size LEB128u */
       ftentry = ftentry->next;
     }
-    WSAACHAR(plines,workbuf,0);			/* End of table */
+    saa_write8(plines,0);			/* End of table */
     linepoff = plines->datalen;
     linelen = linepoff + totlen + 10;
     linebuf = pbuf = nasm_malloc(linelen);
@@ -2157,9 +2148,9 @@ void dwarf64_generate(void)
     psect = dwarf_fsect;
     for (indx = 0; indx < dwarf_nsections; indx++)
     {
-         WSAADLONG(plinesrel,workbuf, linepoff);
-         WSAADLONG(plinesrel,workbuf, ((uint64_t) (psect->section + 2) << 32) +  R_X86_64_64);
-         WSAADLONG(plinesrel,workbuf, (uint64_t) 0);
+         saa_write64(plinesrel, linepoff);
+         saa_write64(plinesrel, ((uint64_t) (psect->section + 2) << 32) +  R_X86_64_64);
+         saa_write64(plinesrel, (uint64_t) 0);
          plinep = psect->psaa;
          saalen = plinep->datalen;
          saa_rnbytes(plinep, pbuf, saalen);
@@ -2294,10 +2285,10 @@ void dwarf64_findsect(const int index)
      dwarf_csect->section = index;
      dwarf_csect->next = 0;
      /* set relocatable address at start of line program */
-     WSAACHAR(plinep,workbuf,DW_LNS_extended_op);
-     WSAACHAR(plinep,workbuf,9);			/* operand length */
-     WSAACHAR(plinep,workbuf,DW_LNE_set_address);
-     WSAADLONG(plinep,workbuf,0);		/* Start Address */
+     saa_write8(plinep,DW_LNS_extended_op);
+     saa_write8(plinep,9);			/* operand length */
+     saa_write8(plinep,DW_LNE_set_address);
+     saa_write64(plinep,0);		/* Start Address */
      /* if first entry */
      if (!dwarf_fsect)
      {
